using JosephsonCircuits
using Test

@testset verbose=true "spiceraw" begin

    @testset "spice_raw_load" begin

        filepath = joinpath(dirname(Base.source_path()),"spiceraw","test01.raw")

        out1 = JosephsonCircuits.spice_raw_load(filepath)

        out2 = JosephsonCircuits.SpiceRaw{Matrix{ComplexF64}}(JosephsonCircuits.SpiceRawHeader("CKT1", "Thu Dec 29 01:29:27 2022", "A.C. Small signal analysis", "complex", 4, 3, "version 4.3.14", ""), Dict("V" => ["v(1)", "v(2)", "v(3)"], "Hz" => ["frequency"]), Dict{String, Matrix{ComplexF64}}("V" => [48.87562301047733 - 7.413126995337487im 49.97131616467212 + 1.1949290155299537im 49.02611690128596 - 6.90980805243651im; -10.116167243319213 + 1.534380793728424im 57.578470543293086 + 1.3775359827006193im 12.368446655904192 - 1.743197747303436im; 0.0 + 0.0im 0.0 + 0.0im 0.0 + 0.0im], "Hz" => [4.0e9 + 0.0im 5.0e9 + 0.0im 6.0e9 + 0.0im]))

        @test out1.variables == out2.variables
        @test out1.values == out2.values
    end

    @testset "spice_raw_load option line" begin

        filepath = joinpath(dirname(Base.source_path()),"spiceraw","test02.raw")

        out1 = JosephsonCircuits.spice_raw_load(filepath)

        out2 = JosephsonCircuits.SpiceRaw{Matrix{ComplexF64}}(JosephsonCircuits.SpiceRawHeader("CKT1", "Thu Dec 29 01:29:27 2022", "A.C. Small signal analysis", "complex", 4, 3, "version 4.3.14", "1"), Dict("V" => ["v(1)", "v(2)", "v(3)"], "Hz" => ["frequency"]), Dict{String, Matrix{ComplexF64}}("V" => [48.87562301047733 - 7.413126995337487im 49.97131616467212 + 1.1949290155299537im 49.02611690128596 - 6.90980805243651im; -10.116167243319213 + 1.534380793728424im 57.578470543293086 + 1.3775359827006193im 12.368446655904192 - 1.743197747303436im; 0.0 + 0.0im 0.0 + 0.0im 0.0 + 0.0im], "Hz" => [4.0e9 + 0.0im 5.0e9 + 0.0im 6.0e9 + 0.0im]))

        @test out1.variables == out2.variables
        @test out1.values == out2.values
    end

    @testset "spice_raw_load extra variables" begin

        filepath = joinpath(dirname(Base.source_path()),"spiceraw","test03.raw")

        @test_warn(
            "Variable line has additional parameters which we are ignoring.",
            JosephsonCircuits.spice_raw_load(filepath)
        )

    end

    @testset "spice_raw_load ascii" begin

        filepath = joinpath(dirname(Base.source_path()),"spiceraw","test04.raw")

        @test_throws(
            ArgumentError("This function only handles Binary files not ASCII."),
            JosephsonCircuits.spice_raw_load(filepath)
        )

    end

    @testset "spice_raw_load flags" begin

        filepath = joinpath(dirname(Base.source_path()),"spiceraw","test05.raw")

        @test_throws(
            ArgumentError("Unknown flag."),
            JosephsonCircuits.spice_raw_load(filepath)
        )

    end

    @testset "spice_raw_load incorrect format" begin

        filepath = joinpath(dirname(Base.source_path()),"spiceraw","test06.raw")

        @test_throws(
            ArgumentError("Line doesn't have the correct format."),
            JosephsonCircuits.spice_raw_load(filepath)
        )

    end

    @testset "spice_raw_load time domain" begin

        filepath = joinpath(dirname(Base.source_path()),"spiceraw","test_td.raw")

        out1 = JosephsonCircuits.spice_raw_load(filepath)

        out2 = JosephsonCircuits.SpiceRaw{Matrix{Float64}}(JosephsonCircuits.SpiceRawHeader("CKT1", "Wed Jan  4 21:57:37 2023", "Transient analysis", "real", 4, 381, "version 4.3.17", ""), Dict("S" => ["time"], "V" => ["v(1)", "v(2)", "v(3)"]), Dict("S" => [0.0 2.6315734072138795e-12 5.263146814427759e-12 7.894720221641638e-12 1.0526293628855518e-11 1.3157867036069397e-11 1.5789440443283277e-11 1.8421013850497156e-11 2.1052587257711036e-11 2.3684160664924915e-11 2.6315734072138795e-11 2.8947307479352674e-11 3.1578880886566554e-11 3.4210454293780433e-11 3.684202770099431e-11 3.947360110820819e-11 4.210517451542207e-11 4.473674792263595e-11 4.736832132984983e-11 4.999989473706371e-11 5.263146814427759e-11 5.526304155149147e-11 5.789461495870535e-11 6.052618836591923e-11 6.315776177313311e-11 6.578933518034698e-11 6.842090858756085e-11 7.105248199477473e-11 7.36840554019886e-11 7.631562880920247e-11 7.894720221641635e-11 8.157877562363022e-11 8.421034903084409e-11 8.684192243805797e-11 8.947349584527184e-11 9.210506925248571e-11 9.473664265969958e-11 9.736821606691346e-11 9.999978947412733e-11 1.026313628813412e-10 1.0526293628855508e-10 1.0789450969576895e-10 1.1052608310298282e-10 1.131576565101967e-10 1.1578922991741057e-10 1.1842080332462444e-10 1.2105237673183831e-10 1.236839501390522e-10 1.2631552354626606e-10 1.2894709695347993e-10 1.315786703606938e-10 1.3421024376790768e-10 1.3684181717512155e-10 1.3947339058233543e-10 1.421049639895493e-10 1.4473653739676317e-10 1.4736811080397704e-10 1.4999968421119092e-10 1.526312576184048e-10 1.5526283102561866e-10 1.5789440443283254e-10 1.605259778400464e-10 1.6315755124726028e-10 1.6578912465447416e-10 1.6842069806168803e-10 1.710522714689019e-10 1.7368384487611577e-10 1.7631541828332965e-10 1.7894699169054352e-10 1.815785650977574e-10 1.8421013850497127e-10 1.8684171191218514e-10 1.8947328531939901e-10 1.9210485872661289e-10 1.9473643213382676e-10 1.9736800554104063e-10 1.999995789482545e-10 2.0263115235546838e-10 2.0526272576268225e-10 2.0789429916989612e-10 2.1052587257711e-10 2.1315744598432387e-10 2.1578901939153774e-10 2.1842059279875162e-10 2.210521662059655e-10 2.2368373961317936e-10 2.2631531302039324e-10 2.289468864276071e-10 2.3157845983482098e-10 2.342100332420349e-10 2.368416066492488e-10 2.394731800564627e-10 2.421047534636766e-10 2.447363268708905e-10 2.473679002781044e-10 2.4999947368531827e-10 2.5263104709253217e-10 2.5526262049974607e-10 2.5789419390695997e-10 2.6052576731417387e-10 2.6315734072138777e-10 2.6578891412860167e-10 2.6842048753581557e-10 2.7105206094302946e-10 2.7368363435024336e-10 2.7631520775745726e-10 2.7894678116467116e-10 2.8157835457188506e-10 2.8420992797909896e-10 2.8684150138631286e-10 2.8947307479352676e-10 2.9210464820074066e-10 2.9473622160795455e-10 2.9736779501516845e-10 2.9999936842238235e-10 3.0263094182959625e-10 3.0526251523681015e-10 3.0789408864402405e-10 3.1052566205123795e-10 3.1315723545845185e-10 3.1578880886566575e-10 3.1842038227287964e-10 3.2105195568009354e-10 3.2368352908730744e-10 3.2631510249452134e-10 3.2894667590173524e-10 3.3157824930894914e-10 3.3420982271616304e-10 3.3684139612337694e-10 3.3947296953059084e-10 3.4210454293780473e-10 3.4473611634501863e-10 3.4736768975223253e-10 3.4999926315944643e-10 3.5263083656666033e-10 3.5526240997387423e-10 3.5789398338108813e-10 3.6052555678830203e-10 3.631571301955159e-10 3.657887036027298e-10 3.684202770099437e-10 3.710518504171576e-10 3.736834238243715e-10 3.763149972315854e-10 3.789465706387993e-10 3.815781440460132e-10 3.842097174532271e-10 3.86841290860441e-10 3.894728642676549e-10 3.921044376748688e-10 3.947360110820827e-10 3.973675844892966e-10 3.999991578965105e-10 4.026307313037244e-10 4.052623047109383e-10 4.078938781181522e-10 4.105254515253661e-10 4.1315702493258e-10 4.157885983397939e-10 4.184201717470078e-10 4.210517451542217e-10 4.236833185614356e-10 4.263148919686495e-10 4.289464653758634e-10 4.315780387830773e-10 4.342096121902912e-10 4.368411855975051e-10 4.39472759004719e-10 4.421043324119329e-10 4.447359058191468e-10 4.473674792263607e-10 4.499990526335746e-10 4.526306260407885e-10 4.552621994480024e-10 4.578937728552163e-10 4.605253462624302e-10 4.631569196696441e-10 4.657884930768579e-10 4.684200664840718e-10 4.710516398912857e-10 4.736832132984996e-10 4.763147867057135e-10 4.789463601129274e-10 4.815779335201413e-10 4.842095069273552e-10 4.868410803345691e-10 4.89472653741783e-10 4.921042271489969e-10 4.947358005562108e-10 4.973673739634247e-10 4.999989473706386e-10 5.026305207778525e-10 5.052620941850664e-10 5.078936675922803e-10 5.105252409994942e-10 5.131568144067081e-10 5.15788387813922e-10 5.184199612211359e-10 5.210515346283498e-10 5.236831080355637e-10 5.263146814427776e-10 5.289462548499915e-10 5.315778282572054e-10 5.342094016644193e-10 5.368409750716332e-10 5.394725484788471e-10 5.42104121886061e-10 5.447356952932749e-10 5.473672687004888e-10 5.499988421077027e-10 5.526304155149166e-10 5.552619889221305e-10 5.578935623293444e-10 5.605251357365583e-10 5.631567091437722e-10 5.657882825509861e-10 5.684198559582e-10 5.710514293654139e-10 5.736830027726278e-10 5.763145761798417e-10 5.789461495870556e-10 5.815777229942695e-10 5.842092964014834e-10 5.868408698086973e-10 5.894724432159112e-10 5.921040166231251e-10 5.94735590030339e-10 5.973671634375529e-10 5.999987368447668e-10 6.026303102519807e-10 6.052618836591946e-10 6.078934570664085e-10 6.105250304736224e-10 6.131566038808363e-10 6.157881772880502e-10 6.184197506952641e-10 6.21051324102478e-10 6.236828975096919e-10 6.263144709169058e-10 6.289460443241197e-10 6.315776177313336e-10 6.342091911385475e-10 6.368407645457614e-10 6.394723379529753e-10 6.421039113601892e-10 6.447354847674031e-10 6.47367058174617e-10 6.499986315818309e-10 6.526302049890448e-10 6.552617783962586e-10 6.578933518034725e-10 6.605249252106864e-10 6.631564986179003e-10 6.657880720251142e-10 6.684196454323281e-10 6.71051218839542e-10 6.736827922467559e-10 6.763143656539698e-10 6.789459390611837e-10 6.815775124683976e-10 6.842090858756115e-10 6.868406592828254e-10 6.894722326900393e-10 6.921038060972532e-10 6.947353795044671e-10 6.97366952911681e-10 6.999985263188949e-10 7.026300997261088e-10 7.052616731333227e-10 7.078932465405366e-10 7.105248199477505e-10 7.131563933549644e-10 7.157879667621783e-10 7.184195401693922e-10 7.210511135766061e-10 7.2368268698382e-10 7.263142603910339e-10 7.289458337982478e-10 7.315774072054617e-10 7.342089806126756e-10 7.368405540198895e-10 7.394721274271034e-10 7.421037008343173e-10 7.447352742415312e-10 7.473668476487451e-10 7.49998421055959e-10 7.526299944631729e-10 7.552615678703868e-10 7.578931412776007e-10 7.605247146848146e-10 7.631562880920285e-10 7.657878614992424e-10 7.684194349064563e-10 7.710510083136702e-10 7.736825817208841e-10 7.76314155128098e-10 7.789457285353119e-10 7.815773019425258e-10 7.842088753497397e-10 7.868404487569536e-10 7.894720221641675e-10 7.921035955713814e-10 7.947351689785953e-10 7.973667423858092e-10 7.999983157930231e-10 8.02629889200237e-10 8.052614626074509e-10 8.078930360146648e-10 8.105246094218787e-10 8.131561828290926e-10 8.157877562363065e-10 8.184193296435204e-10 8.210509030507343e-10 8.236824764579482e-10 8.263140498651621e-10 8.28945623272376e-10 8.315771966795899e-10 8.342087700868038e-10 8.368403434940177e-10 8.394719169012316e-10 8.421034903084455e-10 8.447350637156594e-10 8.473666371228733e-10 8.499982105300872e-10 8.526297839373011e-10 8.55261357344515e-10 8.578929307517289e-10 8.605245041589428e-10 8.631560775661567e-10 8.657876509733706e-10 8.684192243805845e-10 8.710507977877984e-10 8.736823711950123e-10 8.763139446022262e-10 8.789455180094401e-10 8.81577091416654e-10 8.842086648238679e-10 8.868402382310818e-10 8.894718116382956e-10 8.921033850455095e-10 8.947349584527234e-10 8.973665318599373e-10 8.999981052671512e-10 9.026296786743651e-10 9.05261252081579e-10 9.078928254887929e-10 9.105243988960068e-10 9.131559723032207e-10 9.157875457104346e-10 9.184191191176485e-10 9.210506925248624e-10 9.236822659320763e-10 9.263138393392902e-10 9.289454127465041e-10 9.31576986153718e-10 9.342085595609317e-10 9.368401329681455e-10 9.394717063753593e-10 9.421032797825731e-10 9.44734853189787e-10 9.473664265970007e-10 9.499980000042145e-10 9.526295734114283e-10 9.55261146818642e-10 9.578927202258559e-10 9.605242936330697e-10 9.631558670402835e-10 9.657874404474973e-10 9.68419013854711e-10 9.710505872619249e-10 9.736821606691387e-10 9.763137340763525e-10 9.789453074835662e-10 9.8157688089078e-10 9.842084542979938e-10 9.868400277052076e-10 9.894716011124214e-10 9.921031745196352e-10 9.94734747926849e-10 9.973663213340628e-10 9.999978947412766e-10], "V" => [2.0335314834290176e-24 2.0067877638938518e-16 2.866656112878581e-15 1.312045759768243e-14 3.7684716292020694e-14 8.399317770769578e-14 1.595444657315368e-13 2.715725223874628e-13 4.266428411788269e-13 6.305583468674491e-13 8.880064024847426e-13 1.2026017611931e-12 1.5765345596937798e-12 2.010714874047631e-12 2.5044229885403255e-12 3.0555312820375165e-12 3.660180412082472e-12 4.313049238387238e-12 5.007093217509339e-12 5.733823655757867e-12 6.4831529937510326e-12 7.243635264281789e-12 8.002468214724697e-12 8.745639446855764e-12 9.458137349799924e-12 1.0123943216728018e-11 1.0726501304348225e-11 1.1248497080119271e-11 1.1672630979063088e-11 1.1981127393980285e-11 1.2156846622379238e-11 1.2182478571585082e-11 1.2042012958389473e-11 1.1719585987061903e-11 1.1201306420361214e-11 1.047376192066826e-11 9.526131778986198e-12 8.348450140296945e-12 6.933842986404909e-12 5.276691060995185e-12 3.374858324904303e-12 1.227839718062735e-12 -1.1611183864296871e-12 -3.786806612872668e-12 -6.640135604061643e-12 -9.709785576920049e-12 -1.2980578862511197e-11 -1.643492391586558e-11 -2.005154900949171e-11 -2.3806686433448948e-11 -2.76732203197818e-11 -3.1621572464740944e-11 -3.561929533346419e-11 -3.963163899534781e-11 -4.362159633834684e-11 -4.755014994717502e-11 -5.1376751768228775e-11 -5.5059268192267605e-11 -5.85548522592428e-11 -6.18196235022862e-11 -6.480986868433298e-11 -6.748150891106473e-11 -6.979154620446019e-11 -7.169738287991932e-11 -7.315841721059413e-11 -7.413528811887497e-11 -7.459151479705435e-11 -7.449274262543642e-11 -7.380831965749566e-11 -7.251061684425597e-11 -7.057643922560157e-11 -6.798648480937712e-11 -6.472650024960687e-11 -6.078692821739005e-11 -5.616373603893967e-11 -5.085828155911765e-11 -4.487776802987114e-11 -3.8235334091649243e-11 -3.095011815719556e-11 -2.304755031870606e-11 -1.4559043238503092e-11 -5.522433903351786e-12 4.01864600052512e-12 1.4014260922277847e-11 2.440903228656865e-11 3.5141673424481106e-11 4.614595764459372e-11 5.7350434706481273e-11 6.867935116057965e-11 8.005270677332456e-11 9.138694694491903e-11 1.0259553172141882e-10 1.1358920254340778e-10 1.2427724238803077e-10 1.3456711092499134e-10 1.4436657060276055e-10 1.5358248688824102e-10 1.6212398077513338e-10 1.699002051614935e-10 1.768246587617913e-10 1.8281177636590516e-10 1.8778250240349377e-10 1.9165957183578655e-10 1.943743940936147e-10 1.958609521219407e-10 1.9606399007337095e-10 1.9493151941456715e-10 1.9242423812850488e-10 1.8850670308450427e-10 1.831578366274681e-10 1.7636090134508413e-10 1.681148743171019e-10 1.5842344122769674e-10 1.4730694213504407e-10 1.3479068340258944e-10 1.2091708486865618e-10 1.0573368682144618e-10 8.930505070587757e-11 7.17009588751546e-11 5.300742972087156e-11 3.331594714892879e-11 1.2732566543354886e-11 -8.630924383050714e-12 -3.065032738273662e-11 -5.3195712680484197e-11 -7.612727638162337e-11 -9.929977724332122e-11 -1.2256094281707107e-10 -1.4575366644067894e-10 -1.6871688084994837e-10 -1.9128560065874023e-10 -2.132940919782122e-10 -2.3457392324276305e-10 -2.5495917710330633e-10 -2.7428274150733997e-10 -2.9238323926964986e-10 -3.090998488072669e-10 -3.242805704227279e-10 -3.3777590944260235e-10 -3.4944805893196777e-10 -3.5916380514897e-10 -3.6680418605478607e-10 -3.7225698858332245e-10 -3.7542644208819236e-10 -3.762256682650975e-10 -3.745859889462133e-10 -3.7044966270295145e-10 -3.6377842667115803e-10 -3.545468097941504e-10 -3.427495872971516e-10 -3.2839590087544315e-10 -3.1151537986188867e-10 -2.9215322583290167e-10 -2.7037484348639756e-10 -2.462619642049649e-10 -2.1991572856231548e-10 -1.9145383398902287e-10 -1.6101211547348985e-10 -1.2874261069826147e-10 -9.481379130485592e-11 -5.940934907268142e-11 -2.2727332239260796e-11 1.502062703613384e-11 5.361092326064226e-11 9.280866105535227e-11 1.3236958441162919e-10 1.7204108060861725e-10 2.1156393746748602e-10 2.506741158279585e-10 2.891038033856297e-10 3.265844203561825e-10 3.6284642263129174e-10 3.976239974087608e-10 4.306530797938241e-10 4.6167816142968643e-10 4.904480246495362e-10 5.16725027006309e-10 5.402781258901748e-10 5.608949159109496e-10 5.78371613607813e-10 5.925280145856405e-10 6.031942304327485e-10 6.102286011722986e-10 6.135011214646134e-10 6.129141952000581e-10 6.083828510690409e-10 5.998580648725992e-10 5.873040401587006e-10 5.707236557975427e-10 5.501332693056531e-10 5.25589678852435e-10 4.971630881476203e-10 4.6496481022885207e-10 4.291192059444831e-10 3.8979120879259865e-10 3.4715820645981e-10 3.0143633881727584e-10 2.5285342611599555e-10 2.0167289291034166e-10 1.481689501533526e-10 9.264693269228033e-11 3.542200908358623e-11 -2.3165360543510934e-11 -8.276561174310783e-11 -1.4301119055575144e-10 -2.0352621229418526e-10 -2.639246804070538e-10 -3.2381362166237577e-10 -3.8279804258768205e-10 -4.4047791506908533e-10 -4.96459275343527e-10 -5.503456224174343e-10 -6.017543818322967e-10 -6.503034367460697e-10 -6.956320293651857e-10 -7.373832691956732e-10 -7.752284344684707e-10 -8.08846389560781e-10 -8.379501808784299e-10 -8.622643427598462e-10 -8.815526595066134e-10 -8.955943447131786e-10 -9.04211864178831e-10 -9.072469376432935e-10 -9.045873851815745e-10 -8.961438244623617e-10 -8.818746132283341e-10 -8.61764002693629e-10 -8.358443965169092e-10 -8.0417657655306e-10 -7.668686784986554e-10 -7.240589359723343e-10 -6.759309775948021e-10 -6.226993464752509e-10 -5.646209446380686e-10 -5.019833854983463e-10 -4.351126402178209e-10 -3.643640793117215e-10 -2.901266021723262e-10 -2.1281602949011503e-10 -1.3287621214340124e-10 -5.077425751454765e-11 3.3000692262892e-11 1.1794107127513828e-10 2.0352358850887543e-10 2.892124694598369e-10 3.7446265235951164e-10 4.5872353713461523e-10 5.414416189306727e-10 6.220656603177612e-10 7.000476710045689e-10 7.748504103978927e-10 8.459456449840017e-10 9.128248295907021e-10 9.74993651397827e-10 1.031986629452442e-9 1.083357086021532e-9 1.1286962470168373e-9 1.1676179646334069e-9 1.1997827003161553e-9 1.2248765524156442e-9 1.242640264100172e-9 1.2528423791432765e-9 1.2553131475033365e-9 1.24991193250524e-9 1.236565596044723e-9 1.2152306008799642e-9 1.1859351562890799e-9 1.148736788169543e-9 1.1037672262194947e-9 1.0511865393267979e-9 9.91229459330312e-10 9.241574865031916e-10 8.503050852683727e-10 7.700314410966283e-10 6.837647425613121e-10 5.919555037247925e-10 4.951169762948635e-10 3.9378212847723477e-10 2.8853812127359667e-10 1.7998914508698454e-10 6.878285433665901e-11 -4.4418702142534503e-11 -1.589227570203025e-10 -2.7402461180750727e-10 -3.890033359128461e-10 -5.031284063184311e-10 -6.156686407346691e-10 -7.258860330645858e-10 -8.330574355620508e-10 -9.3645652104568e-10 -1.0353867208860262e-9 -1.1291527465188427e-9 -1.2171029654854633e-9 -1.2985921688551015e-9 -1.373031354329005e-9 -1.4398436019989206e-9 -1.4985191002425507e-9 -1.5485660925142857e-9 -1.5895689319436911e-9 -1.6211360844333038e-9 -1.6429590321087966e-9 -1.6547592635526142e-9 -1.6563459621442329e-9 -1.6475637862497633e-9 -1.6283474707126173e-9 -1.598672004397085e-9 -1.5586025335673724e-9 -1.5082482899846606e-9 -1.4478065030722806e-9 -1.377521129624856e-9 -1.2977198447789575e-9 -1.2087782869732403e-9 -1.1111495913838883e-9 -1.0053345051316123e-9 -8.91903335325484e-10 -7.714719246034443e-10 -6.447162928917377e-10 -5.123541075169543e-10 -3.7515268674909306e-10 -2.3391526708046755e-10 -8.94833915658025e-11 5.72730161657807e-11 2.0545754979244842e-10 3.541546712737751e-10 5.024343223373069e-10 6.49358110354566e-10 7.939848907469563e-10 9.35377303681075e-10 1.0726064968722932e-9 1.20476055634384e-9 1.330946539226949e-9 1.4503022869303961e-9 1.5619940281186495e-9 1.665232925103795e-9 1.759266687339523e-9 1.8434019845058918e-9 1.9169887762437943e-9 1.9794494343212366e-9 2.0302548372082995e-9 2.068960688072348e-9 2.0951748321931053e-9 2.1086008493974566e-9 2.108996514339657e-9 2.096224295961794e-9 2.0702013788859537e-9 2.0309562352836064e-9 1.9785711415818512e-9 1.9132435180318837e-9 1.8352223887098029e-9 1.7448727303549157e-9 1.6426078145421165e-9 1.5289543896033008e-9 1.4044832910658053e-9 1.2698729178988695e-9 1.1258408806895143e-9 9.732029442301218e-10 8.128087836485569e-10 6.455933658767275e-10 4.72520085355679e-10 2.946214611919658e-10 1.1295301216201427e-10 -7.137983464527504e-11 -2.5725380560569185e-10 -4.4351762346701606e-10 -6.290068484487328e-10 -8.125531689034641e-10 -9.929791638784088e-10 -1.169129161673969e-9 -1.3398424183566389e-9 -1.5040051662372024e-9 -1.6605038180512645e-9 -1.808294946283229e-9 -1.946341930353163e-9 -2.073699166886691e-9 -2.189435346189075e-9 -2.292728290422701e-9 -2.3827781857194e-9 -2.4589093171689475e-9 -2.520476668403232e-9 -2.5669734887286423e-9; 1.848664431838641e-25 1.8239419665165223e-17 2.603707820586884e-16 1.190303311388106e-15 3.413013358662368e-15 7.589937003261897e-15 1.437615502557434e-14 2.4386184876182105e-14 3.815370313586877e-14 5.611914584342759e-14 7.859567185394193e-14 1.05769838338838e-13 1.3767030467812354e-13 1.7417813975769548e-13 2.1499942815471182e-13 2.596824372689974e-13 3.0759732193997224e-13 3.579566482327762e-13 4.0980646113737673e-13 4.620456332071705e-13 5.134331130045167e-13 5.62600680132068e-13 6.080813056731073e-13 6.483093706391714e-13 6.816747153769674e-13 7.065043509421736e-13 7.211459097697449e-13 7.239254347279445e-13 7.132625445089029e-13 6.876000665234128e-13 6.455513060003205e-13 5.857990961429637e-13 5.072731567893139e-13 4.0901852234624184e-13 2.903977931460931e-13 1.509337177744148e-13 -9.476213341861935e-15 -1.9072908759527242e-13 -3.9231043518539357e-13 -6.134584466144397e-13 -8.529822974679765e-13 -1.1094113683848202e-12 -1.3808493111913408e-12 -1.6650998851091954e-12 -1.9595663586580028e-12 -2.261346912962052e-12 -2.5671866225120105e-12 -2.873537212571216e-12 -3.176565978711604e-12 -3.4721768985791332e-12 -3.756077605075798e-12 -4.023761292006392e-12 -4.2706294474947495e-12 -4.491936486629248e-12 -4.682962596707573e-12 -4.838925395941134e-12 -4.9551941624901045e-12 -5.0271748754995035e-12 -5.050554510417419e-12 -5.02116739826752e-12 -4.935256365002129e-12 -4.789329301775471e-12 -4.580422867782537e-12 -4.305958387332924e-12 -3.963993741567866e-12 -3.5530871291732907e-12 -3.0725237072926965e-12 -2.5221943535755886e-12 -1.902785585173897e-12 -1.2156782767323506e-12 -4.63092329390041e-13 3.519925753815255e-13 1.225742291035952e-12 2.153526755826865e-12 3.129875567351836e-12 4.148520942881541e-12 5.202398944474373e-12 6.283685832067149e-12 7.383835308825184e-12 8.493620962005518e-12 9.603195058683117e-12 1.070215302922084e-11 1.1779595023979398e-11 1.2824230907352764e-11 1.3824422325338745e-11 1.4768348163327604e-11 1.564400259369362e-11 1.64394411562792e-11 1.7142709439333736e-11 1.774218866277092e-11 1.822643052497764e-11 1.8584618262541526e-11 1.8806285928148875e-11 1.8881906299843862e-11 1.8802477513598808e-11 1.8560242920890272e-11 1.8148134557884945e-11 1.7560622218047704e-11 1.6793011567049795e-11 1.58424105220763e-11 1.4706889677775946e-11 1.3386543878937584e-11 1.1882533661516864e-11 1.0198209076109967e-11 8.338062623098389e-12 6.30887141869728e-12 4.1186040810125826e-12 1.7775267830604144e-12 -7.028818764318138e-13 -3.3089249160267326e-12 -6.025699557132394e-12 -8.83626322845071e-12 -1.1722500248779327e-11 -1.466453727455653e-11 -1.764138431068449e-11 -2.0630679844450872e-11 -2.3609023021275723e-11 -2.655212759500741e-11 -2.9434764758754474e-11 -3.223138982921525e-11 -3.491565352805076e-11 -3.746149629658582e-11 -3.984225160668028e-11 -4.2032164093819256e-11 -4.4005123432670695e-11 -4.573654909016436e-11 -4.720180607386277e-11 -4.8378385384905147e-11 -4.9244063938672076e-11 -4.9779299052461886e-11 -4.996520297388817e-11 -4.978606251643299e-11 -4.922720348365006e-11 -4.827754327460536e-11 -4.692742177186922e-11 -4.5171095623171995e-11 -4.300460973602516e-11 -4.0428147066817265e-11 -3.74440087717027e-11 -3.4058742575266215e-11 -3.028129007771229e-11 -2.6124829815255846e-11 -2.1605137602031473e-11 -1.674209673002708e-11 -1.15583020779709e-11 -6.080208853646625e-12 -3.369942245342066e-13 5.638663247281479e-12 1.1811468692284601e-11 1.814308706446109e-11 2.459279729233566e-11 3.1117388832638766e-11 3.767162449129903e-11 4.420840212284175e-11 5.0679081169492506e-11 5.7033836439437574e-11 6.322191609525063e-11 6.91921000257053e-11 7.489296607285001e-11 8.027335539531308e-11 8.528272840951397e-11 8.987153451397463e-11 9.399173894754879e-11 9.759699490402419e-11 1.0064341688904697e-10 1.0308945997779036e-10 1.0489700412018009e-10 1.0603085820239069e-10 1.0646020319510247e-10 1.0615765761659388e-10 1.0510110579300573e-10 1.0327228584382481e-10 1.0065900324558772e-10 9.725323183899935e-11 9.30536832366911e-11 8.806344457868656e-11 8.229284221956419e-11 7.575667577727691e-11 6.847728130409583e-11 6.048145810840972e-11 5.1803603754897374e-11 4.2482466546014135e-11 3.2564193825572624e-11 2.2099077129546737e-11 1.1144328058341072e-11 -2.3898685928084477e-13 -1.1983650338634638e-11 -2.4018948554120187e-11 -3.6269077927048084e-11 -4.865515060084228e-11 -6.109453300815586e-11 -7.350195723923556e-11 -8.579001755840197e-11 -9.786914507349677e-11 -1.0964947485486702e-10 -1.2103945535475094e-10 -1.3194927415385947e-10 -1.4228789944922966e-10 -1.5196821652605254e-10 -1.60902336462085e-10 -1.6900854844713577e-10 -1.7620478030268117e-10 -1.8241735659484533e-10 -1.8757284516191932e-10 -1.9160800732726472e-10 -1.9446070547015445e-10 -1.960804900275861e-10 -1.9641893058165146e-10 -1.9544048337951887e-10 -1.9311255691038652e-10 -1.8941631522954877e-10 -1.8433678174651337e-10 -1.7787324827202097e-10 -1.700297036977767e-10 -1.608245459840394e-10 -1.5028159424433961e-10 -1.384388398566761e-10 -1.2534026231108243e-10 -1.1104340478337163e-10 -9.561216752035979e-11 -7.912305065256888e-11 -6.16590458151981e-11 -4.3314452415332415e-11 -2.4189930856921944e-11 -4.395865345963543e-12 1.595141702822696e-11 3.6727696344139214e-11 5.7803528958461935e-11 7.904360450514837e-11 1.003084120930829e-10 1.2145473812567162e-10 1.4233648730935608e-10 1.6280610277204788e-10 1.827147400229189e-10 2.0191434878467643e-10 2.2025742793803193e-10 2.375994789141453e-10 2.537985872882972e-10 2.6871794222748634e-10 2.822254944896427e-10 2.941962364834528e-10 3.0451217623593495e-10 3.130640886976964e-10 3.197520206020119e-10 3.244862556948366e-10 3.2718852678029795e-10 3.2779198485546437e-10 3.262432431072222e-10 3.225012010979498e-10 3.1653998275028645e-10 3.083465397218702e-10 2.979244737695847e-10 2.8529042562810506e-10 2.7047869438354397e-10 2.5353650636261863e-10 2.345292635000955e-10 2.1353487735344882e-10 1.9064939848007527e-10 1.6598068980473296e-10 1.3965411349717096e-10 1.1180590144833544e-10 8.258850889054622e-11 5.216411356838761e-11 2.070919012587524e-11 -1.1591370295009637e-11 -4.4540950540252294e-11 -7.79359329127063e-11 -1.1156416114394265e-10 -1.4520795426980776e-10 -1.7864484637311288e-10 -2.1164833330458639e-10 -2.4399137290818397e-10 -2.754443352819564e-10 -3.05781698216605e-10 -3.347767241480654e-10 -3.6221171286542757e-10 -3.878690070630918e-10 -4.115450676673172e-10 -4.3303751403688454e-10 -4.521631681639116e-10 -4.68740963143355e-10 -4.826139636583355e-10 -4.936281243506284e-10 -5.016581389972921e-10 -5.065821948847557e-10 -5.083111910098669e-10 -5.067605375839401e-10 -5.018807510696427e-10 -4.936285111953009e-10 -4.81996739326819e-10 -4.669862372302542e-10 -4.486342153275759e-10 -4.269872999429037e-10 -4.021277639860938e-10 -3.741485612624007e-10 -3.4317662694652665e-10 -3.0935048709614077e-10 -2.7284013177999406e-10 -2.338276203721328e-10 -1.925231812257031e-10 -1.4914909370560942e-10 -1.0395182908271762e-10 -5.718934414515941e-11 -9.139242335487464e-12 3.9910532986980945e-11 8.965737662212055e-11 1.3979008203386048e-10 1.8998806576316252e-10 2.39924465544354e-10 2.892684811368831e-10 3.376859078309359e-10 3.8484404763628837e-10 4.3041016853939074e-10 4.740583496417587e-10 5.154664236767903e-10 5.54324059132453e-10 5.903287820957708e-10 6.231945705366718e-10 6.526475594583354e-10 6.784344290565973e-10 7.003183735456568e-10 7.180866062230308e-10 7.315471176045476e-10 7.405346955354767e-10 7.449089140837923e-10 7.445581733901455e-10 7.393992500746028e-10 7.293789958315703e-10 7.144756470367227e-10 6.946979782550174e-10 6.700884185033943e-10 6.407196298463881e-10 6.066993165067932e-10 5.681643826642132e-10 5.2528715516029e-10 4.78267460900103e-10 4.2733982160742733e-10 3.727639718437602e-10 3.1483243269587773e-10 2.5386015821934854e-10 1.9019176159672382e-10 1.2419112278826355e-10 5.6247432755334125e-11 -1.323428926774087e-11 -8.382924389205822e-11 -1.5510039674251486e-10 -2.2659756774964067e-10 -2.9786197310966514e-10 -3.684293009597191e-10 -4.3783019297944837e-10 -5.055982029147046e-10 -5.712652182293876e-10 -6.343751025796999e-10 -6.944731656731461e-10 -7.511261490408796e-10 -8.039050094753986e-10 -8.52411721337256e-10 -8.962548429787398e-10 -9.350833822222232e-10 -9.685548614612804e-10 -9.963762315267274e-10 -1.0182644332501616e-9 -1.0339940560497958e-9 -1.0433507034690879e-9 -1.0461847853535437e-9 -1.0423583225232694e-9 -1.0318039455573024e-9 -1.0144661083267641e-9 -9.903638249988846e-10 -9.595287334361817e-10 -9.220676139470714e-10 -8.781018699787966e-10 -8.2782558344795e-10 -7.714496190876769e-10 -7.092531026411371e-10 -6.415334883999467e-10 -5.686504155699636e-10 -4.909826505370339e-10 -4.089635919962576e-10 -3.2304570569719195e-10 -2.3372777082376913e-10; 1.0190375787955931e-23 3.01463641653152e-14 8.656258361285592e-13 6.0208029734452474e-12 2.334814111464772e-11 6.582812920623056e-11 1.5169749229406956e-10 3.0436279436521655e-10 5.517246984313628e-10 9.257643288406677e-10 1.4612455151901929e-9 2.1952649055875865e-9 3.1653662048827113e-9 4.409263889051943e-9 5.962385997891061e-9 7.857937276192743e-9 1.0123923461042727e-8 1.2783668504093306e-8 1.5852438003011502e-8 1.9338454316399334e-8 2.3239319025525754e-8 2.7543478433798485e-8 3.2226737169655584e-8 3.725401929412461e-8 4.257635965862996e-8 4.813269026175242e-8 5.384778406713773e-8 5.963368206413744e-8 6.538913637203101e-8 7.100019170020512e-8 7.634172076128705e-8 8.127659610480886e-8 8.565992142223649e-8 8.933618744199477e-8 9.214674966478566e-8 9.392446160336213e-8 9.450462635675944e-8 9.371707887988001e-8 9.140017253768902e-8 8.739073586910751e-8 8.154051395976503e-8 7.370434226015426e-8 6.375841631175972e-8 5.158712700574501e-8 3.710238134990897e-8 2.0229645569243283e-8 9.274314536077881e-10 -2.082684008991818e-8 -4.501812846200976e-8 -7.16074530737746e-8 -1.0051489695088053e-7 -1.3163221621247285e-7 -1.6480843140789147e-7 -1.9986082342519107e-7 -2.365638601746791e-7 -2.746581504708378e-7 -3.138432715135476e-7 -3.53784405966161e-7 -3.941094061293107e-7 -4.3441302270339886e-7 -4.7425827709595487e-7 -5.131783911598965e-7 -5.50682255348245e-7 -5.862543727444139e-7 -6.193639257903273e-7 -6.494632393435465e-7 -6.75999681934127e-7 -6.984133149502888e-7 -7.161506357077474e-7 -7.28662197111862e-7 -7.354170283843521e-7 -7.35901067028354e-7 -7.29630996050352e-7 -7.161543155817285e-7 -6.950613269600387e-7 -6.659875761991329e-7 -6.286228052848754e-7 -5.82716319410221e-7 -5.280819191298855e-7 -4.646064784703364e-7 -3.9225016719213397e-7 -3.1105818361651156e-7 -2.2115596689612548e-7 -1.2276361844261245e-7 -1.6186276086758447e-8 9.81697012504445e-8 2.1981049233600266e-7 3.4813827693825145e-7 4.824678535895338e-7 6.220116473960287e-7 7.658973393487599e-7 9.131567690615016e-7 1.062742184614412e-6 1.2135217801358039e-6 1.364291698454869e-6 1.5137812346378104e-6 1.6606573515369346e-6 1.8035427403493621e-6 1.941009377933297e-6 2.071612633944363e-6 2.193870341833726e-6 2.30631595345837e-6 2.407459741104759e-6 2.495863605351182e-6 2.5700813943110663e-6 2.6287573576645113e-6 2.670543173772569e-6 2.6942212103877806e-6 2.69859669218398e-6 2.6826458855228386e-6 2.6453829085866368e-6 2.5860317239267014e-6 2.5038683688410075e-6 2.3984142471681613e-6 2.2692560043834193e-6 2.1162561026485424e-6 1.939354136674402e-6 1.7387889161766534e-6 1.5148876140409505e-6 1.2682892433330123e-6 9.997358660766582e-7 7.10277640731439e-7 4.010841030459783e-7 7.36185559110583e-8 -2.7052410143264105e-7 -6.294788704284157e-7 -1.001242062429406e-6 -1.3835725067536602e-6 -1.774089267674671e-6 -2.170214802261879e-6 -2.5692374678319502e-6 -2.9682971680584613e-6 -3.3644126152730233e-6 -3.754508261120833e-6 -4.1354077600911615e-6 -4.503899274637445e-6 -4.856697997856605e-6 -5.1905453318540895e-6 -5.502144598187212e-6 -5.788288189887862e-6 -6.045771661302394e-6 -6.271541820818841e-6 -6.4625951870074964e-6 -6.61613918420547e-6 -6.729481429161777e-6 -6.800195766680806e-6 -6.826008972721822e-6 -6.8049633588773435e-6 -6.735307221532833e-6 -6.615646171347724e-6 -6.444843018030494e-6 -6.222150844407241e-6 -5.9471270051571405e-6 -5.619742244543297e-6 -5.2403121363188025e-6 -4.8095781993639615e-6 -4.328658464582092e-6 -3.7990985132734135e-6 -3.2228409922278653e-6 -2.6022467069802704e-6 -1.940080900997349e-6 -1.2395068851921473e-6 -5.040848464418242e-7 2.6225712191882e-7 1.0552179082285441e-6 1.8701677858163554e-6 2.7021450452910335e-6 3.545906570455171e-6 4.395935802906477e-6 5.246488459194964e-6 6.09162256832426e-6 6.92522697037903e-6 7.741084833284788e-6 8.532871791434397e-6 9.294264442549338e-6 1.0018894838735547e-5 1.0700514921433669e-5 1.1332894634655595e-5 1.1910052032627215e-5 1.2426083588047041e-5 1.2875467786450383e-5 1.3252818264390474e-5 1.355326609949471e-5 1.3772129215801388e-5 1.3905375332858751e-5 1.3949204565796502e-5 1.3900591186511779e-5 1.3756780438157247e-5 1.3515902993780503e-5 1.317639157939199e-5 1.2737657427349593e-5 1.2199437208777369e-5 1.156251603404892e-5 1.082802025654688e-5 9.99816685652132e-6 9.075522694164151e-6 8.063755628256622e-6 6.966885551249215e-6 5.790008552827711e-6 4.538569338125537e-6 3.2190263634998244e-6 1.8381787952322421e-6 4.0373553418569925e-7 -1.0762550245662228e-6 -2.592991170817613e-6 -4.137332598043512e-6 -5.699569152608298e-6 -7.269670827076528e-6 -8.837238651038557e-6 -1.039159188029387e-5 -1.192189361793964e-5 -1.3417082828602474e-5 -1.4866161319995505e-5 -1.6257983374840783e-5 -1.7581685343493043e-5 -1.8826350232934884e-5 -1.9981556576948315e-5 -2.103693906386607e-5 -2.1982829752463635e-5 -2.2809738598827665e-5 -2.3509057629571808e-5 -2.4072486919633085e-5 -2.4492771134108125e-5 -2.4763097030915115e-5 -2.48778319059179e-5 -2.4831918011151495e-5 -2.4621583901137138e-5 -2.424375928170749e-5 -2.369673302113946e-5 -2.297960893962535e-5 -2.2092888225727837e-5 -2.103798265474457e-5 -1.9817704195358e-5 -1.843584726582501e-5 -1.68975742352355e-5 -1.5209072791126184e-5 -1.3377832599161908e-5 -1.1412377901760071e-5 -9.322437528475074e-6 -7.118746737673187e-6 -4.813119770329788e-6 -2.418308969229579e-6 5.200417158119635e-8 2.583310131862973e-6 5.160339666207141e-6 7.767147542375031e-6 1.0387212704092618e-5 1.3003530857342091e-5 1.5598714826401307e-5 1.815512486146426e-5 2.065493474984862e-5 2.308032992284551e-5 2.54135037353418e-5 2.763695306159474e-5 2.9733369784127633e-5 3.168606148629074e-5 3.3478705686416265e-5 3.509591942363012e-5 3.652284815589475e-5 3.774590252884341e-5 3.87521600705311e-5 3.953028033840175e-5 4.0069705455224875e-5 4.03617563273324e-5 4.0398627463774354e-5 4.017465730119133e-5 3.9685123346164666e-5 3.892766914563719e-5 3.790091723776451e-5 3.6606024346375395e-5 3.504514150715997e-5 3.322305806198548e-5 3.1145550187542885e-5 2.8821063676162752e-5 2.62590034436883e-5 2.3471395382600466e-5 2.047117975981519e-5 1.7273784092766984e-5 1.3895492163046745e-5 1.035485335880609e-5 6.671205844469136e-6 2.8658435622833537e-6 -1.0392235305725673e-6 -5.020425224778099e-6 -9.053549298895142e-6 -1.3113299179706123e-5 -1.7173807612269252e-5 -2.1208665797277758e-5 -2.5190981138079968e-5 -2.9093884783100844e-5 -3.28901487474972e-5 -3.655313263677395e-5 -4.005599998203799e-5 -4.337304873991928e-5 -4.647857848170285e-5 -4.934853865428062e-5 -5.195910715341801e-5 -5.428858238791494e-5 -5.6315740087788944e-5 -5.802189022541848e-5 -5.9389082439492255e-5 -6.0402247013411484e-5 -6.104731991768064e-5 -6.131338839411438e-5 -6.119080432444395e-5 -6.067325946731031e-5 -5.9755950752351427e-5 -5.8437517695040404e-5 -5.671831550647711e-5 -5.46021573831316e-5 -5.209474133037017e-5 -4.920515249709537e-5 -4.5944477747831414e-5 -4.232703749777547e-5 -3.8369208827710656e-5 -3.409037180508191e-5 -2.9511947876724135e-5 -2.465806180748897e-5 -1.955478781197909e-5 -1.4230544336137616e-5 -8.71552652902758e-6 -3.041866451437501e-6 2.756782328816636e-6 8.645272567418901e-6 1.4587272010434479e-5 2.0545420827708063e-5 2.6481587889706892e-5 3.235709009095956e-5 3.813295685093687e-5 4.37701374586344e-5 4.922983758475829e-5 5.44736381323135e-5 5.946396734018443e-5 6.416405575362613e-5 6.853860375001987e-5 7.255350227428512e-5 7.617674830134601e-5 7.937786989248499e-5 8.21291313449889e-5 8.440461373970196e-5 8.618173758146646e-5 8.743996599438057e-5 8.816265593477877e-5 8.833537006854833e-5 8.794804868352395e-5 8.699293743400147e-5 8.546705155409846e-5 8.33697493377212e-5 8.070543892308144e-5 7.74808507506454e-5 7.370791589468284e-5 6.940081389127112e-5 6.457893123014798e-5 5.926378290876174e-5 5.3481941182796214e-5 4.726194900727299e-5 4.063709338176683e-5 3.364244553463363e-5 2.6317341049148326e-5 1.8702694338705406e-5 1.0843039984255977e-5 2.7842769363486836e-6 -5.424877197743516e-6 -1.3734554378922476e-5 -2.2093014261221673e-5 -3.0447570581155757e-5 -3.8744738206140564e-5 -4.6930273335455106e-5 -5.4950303559714286e-5 -6.27503911619091e-5 -7.027765742142564e-5 -7.747891183228361e-5 -8.430364861485723e-5 -9.070137651557287e-5 -9.662533220066532e-5 -0.00010202914586747525 -0.00010687111264644606 -0.00011111034019877957 -0.00011471140892622291 -0.00011764015548663761 -0.0001198686729632148]))

        @test out1.variables == out2.variables
        @test out1.values == out2.values

    end

    @testset "parsespicevariable no match" begin

        @test_throws(
            ArgumentError("No match found."),
            JosephsonCircuits.parsespicevariable("")
        )

    end

end